!
! Copyright (c) 2000-2013,2017 Yoshihide Yoshimoto
!
#include "fftgrid.h"

module pot_ldaxc
  private
  public ldaxc
contains
  subroutine ldaxc(nspin,nrx2,nry2,nrz2,nlx,nly,nlz,nlxyz,nkfxa,kf, &
       rhop,rho,va,bb,exa,exc,evxc,excor,evxcor,fs,xctp,mssg,comm)
    !
    ! input
    ! rhop: partial core charge + valence charge (in real space)
    ! rho: valence charge (in real space)
    !
    ! output
    ! rho: Vxc(R)
    !
    use fft_3d
    use timeprof
    use mpi
    implicit none
    integer,intent(in):: nspin,nrx2,nry2,nrz2,nlx,nly,nlz,nlxyz
    integer,intent(in):: nkfxa,kf(3,nkfxa)
    !
    ! input rhop and rho must be padded by zero.
    !
    real(kind=8):: rhop(nlxyz,nspin)
    real(kind=8):: rho(nlxyz,nspin),va,bb(3,3),exa,exc,evxc(0:6)
    real(kind=8):: excor(nlxyz),evxcor(nlxyz,0:6)
    type (fft3_struct):: fs
    integer,intent(in):: xctp
    integer,intent(in):: mssg, comm
    !
    real(kind=8):: minneg
    integer:: count
    integer:: is,ierr
    !
    ! make sure padding
    !
    do is=1,nspin
       call pad_zero(nrx2,nry2,nrz2,nlx,nly,nlz,rhop(1,is))
    end do
    do is=1, nspin
       call fcpchr(nrx2,nry2,nrz2,nlx,nly,nlz,rhop(1,is),minneg,count)
       if (mssg.gt.0) then
          if (nspin .eq. 2) then
             write(6,'(1x,a,i1)') 'SPIN = ', is
          end if
          write(6,'(1x,a)') 'POSITIVE CHARGE FORCED ON RHOV + RHOP IN LDAXC'
          write(6,'(2x,i9,a,i9,a)') count,' of',nrx2*nry2*nrz2, &
               ' negative point found.'
          write(6,'(2x,a,e18.11)') 'Minimum value is ', minneg
          write(6,600) sum(rhop(:,is))*va/(nrx2*nry2*nrz2)
       end if
600    format(2x,'CORRECTED TOTAL CHARGE ',f25.20)
    end do
    if (nspin .eq. 1) then
       call evalxc(fs,nrx2,nry2,nrz2,nlx,nly,nlz,nlxyz,nkfxa,kf, &
            rhop,rho,va,bb,exa,exc,evxc,excor,evxcor,xctp,mssg,comm)
    else if (nspin .eq. 2) then
       call evalxcsp(fs,nrx2,nry2,nrz2,nlx,nly,nlz,nlxyz,nkfxa,kf, &
            rhop,rho,va,bb,exa,exc,evxc,excor,evxcor,xctp,mssg,comm)
    else
       write(6,*) 'invalid value in nspin = ', nspin
       stop
    end if
#ifdef INCONSISTENT_THREAD_COMP
    call tpf_begin(tpf_mpi)
    call MPI_BCAST(evxc,7,MPI_DOUBLE_PRECISION,0,comm,ierr)
    call MPI_BCAST(exa,1,MPI_DOUBLE_PRECISION,0,comm,ierr)
    call MPI_BCAST(exc,1,MPI_DOUBLE_PRECISION,0,comm,ierr)
    call tpf_end(tpf_mpi)
#endif
    return
  end subroutine ldaxc

  subroutine fcpchr(nrx2,nry2,nrz2,nlx,nly,nlz,rho,minneg,count)
    implicit none
    integer,intent(in):: nlx,nly,nlz,nrx2,nry2,nrz2
    real(kind=8):: rho(nlx,nly,nlz),minneg
    integer:: count
    integer:: i,j,k
    minneg = rho(1,1,1)
    count = 0
    do k=1,nrz2
       do j=1,nry2
          do i=1,nrx2
             minneg = min(rho(i,j,k),minneg)
             if (rho(i,j,k) .le. 0.0d0) then
                rho(i,j,k) = 0.0d0
                count = count + 1
             end if
          end do
       end do
    end do
    return
  end subroutine fcpchr

  subroutine evalxc(fs,nrx2,nry2,nrz2,nlx,nly,nlz,nlxyz,nkfxa,kf, &
       rhop,rho,va,bb,exa,exc,evxc,excor,evxcor,xctp,mssg,comm)
    use fft_3d
    use ldapw92_xc
    use mpi
    use timeprof
    use xcf_types
    implicit none
    !
    ! input
    !
    type (fft3_struct):: fs
    integer,intent(in):: nrx2,nry2,nrz2,nlx,nly,nlz,nlxyz
    integer,intent(in):: nkfxa,kf(3,nkfxa)
    real(kind=8):: rhop(nlxyz),rho(nlxyz),va,bb(3,3)
    !
    integer,intent(in):: xctp
    integer,intent(in):: mssg,comm
    !
    ! output
    !
    ! rho contains Vxc
    real(kind=8):: exa,exc,evxc(0:6)
    real(kind=8):: excor(nlxyz),evxcor(nlxyz,0:6)
    !
    integer:: k
    !
    real(kind=8),dimension(:),allocatable:: fxcn
    real(kind=8),allocatable:: pepsxc(:),pfxcn(:)
    integer,allocatable:: r_count(:),r_displ(:)
    integer:: myrank,nproc,irnk,blxyz,pnlxyz
    real(kind=8):: pexc
    integer:: ierr
    !
    call MPI_COMM_SIZE(comm,nproc,ierr)
    call MPI_COMM_RANK(comm,myrank,ierr)
    allocate(r_count(0:nproc-1),r_displ(0:nproc-1),stat=ierr)
    call chkalloc('evalxc','ptmp1',ierr)
    r_count = nlxyz/nproc
    do irnk = 0, mod(nlxyz,nproc)-1
       r_count(irnk) = r_count(irnk)+1
    end do
    r_displ(0) = 0
    do irnk = 0, nproc-2
       r_displ(irnk+1) = r_displ(irnk)+r_count(irnk)
    end do
    blxyz = r_displ(myrank)+1
    pnlxyz = r_count(myrank)
    allocate(pepsxc(pnlxyz),pfxcn(pnlxyz),stat=ierr)
    call chkalloc('evalxc','ptmp2',ierr)
    !
    select case(xctp)
    case(xcf_ldapw92)
       call ldapw92_exchcor(pnlxyz,rhop(blxyz),pepsxc,pfxcn,mssg)
    case default
       if (mssg.gt.0) then
          write(6,*) 'evalxc(): unsupported value in xctp = ', xctp
       end if
       call  MPI_ABORT(MPI_COMM_WORLD,1,ierr)
       stop
    end select
    !
    ! padding: at least rhop is zero
    !
    pexc = 0.0d0
    do k=1,pnlxyz
       pexc = pexc + rhop(r_displ(myrank)+k)*pepsxc(k)
    end do
    call tpf_begin(tpf_mpi)
    call MPI_ALLREDUCE(pexc,exc,1,MPI_DOUBLE_PRECISION,MPI_SUM, &
         comm,ierr)
    !
    ! using pepsxc as a temporal array
    !
    do k=1,pnlxyz
       pepsxc(k) = rhop(r_displ(myrank)+k)*pepsxc(k)
    end do
    call MPI_ALLGATHERV(pepsxc,pnlxyz,MPI_DOUBLE_PRECISION, &
         excor,r_count,r_displ,MPI_DOUBLE_PRECISION,comm,ierr)
    call tpf_end(tpf_mpi)
    deallocate(pepsxc)

    evxc(1:6) = 0.0d0
    !
    allocate(fxcn(nlxyz),stat=ierr)
    call chkalloc('evalxc','tmp3',ierr)
    call tpf_begin(tpf_mpi)
    call MPI_ALLGATHERV(pfxcn,pnlxyz,MPI_DOUBLE_PRECISION, &
         fxcn,r_count,r_displ,MPI_DOUBLE_PRECISION,comm,ierr)
    call tpf_end(tpf_mpi)
    deallocate(pfxcn,r_count,r_displ)
    !
    call pad_zero(nrx2,nry2,nrz2,nlx,nly,nlz,fxcn)
    !
    ! generate Vxc(R) in rho

    call genvxc(nrx2*nry2*nrz2,nlxyz,fxcn,rhop,rho,va, &
         exa,exc,evxc,evxcor(:,0))
    deallocate(fxcn)
    return
  end subroutine evalxc

  subroutine evalxcsp(fs,nrx2,nry2,nrz2,nlx,nly,nlz,nlxyz,nkfxa,kf, &
       rhop,rho,va,bb,exa,exc,evxc,excor,evxcor,xctp,mssg,comm)
    use fft_3d
    use ldapw92_xc
    use mpi
    use timeprof
    use xcf_types
    implicit none
    !
    ! input
    !
    type (fft3_struct):: fs
    integer,intent(in):: nrx2,nry2,nrz2,nlx,nly,nlz,nlxyz
    integer,intent(in):: nkfxa,kf(3,nkfxa)
    real(kind=8):: rhop(nlxyz,2),rho(nlxyz,2),va,bb(3,3)
    !
    integer,intent(in):: xctp
    integer,intent(in):: mssg,comm
    !
    ! output
    !
    !  rho contains Vxc
    real(kind=8):: exa,exc,evxc(0:6)
    real(kind=8):: excor(nlxyz),evxcor(nlxyz,0:6)
    !
    integer:: k,j,is
    !
    real(kind=8),allocatable:: pfxcn(:,:)
    real(kind=8),allocatable:: prhop(:,:)
    real(kind=8),allocatable:: ppfxcn(:,:)
    real(kind=8),allocatable:: pexcor(:)
    integer,allocatable:: r_count(:),r_displ(:)
    integer:: myrank,nproc,irnk,pnlxyz
    integer:: ierr
    real(kind=8):: pexc
    !
    call MPI_COMM_SIZE(comm,nproc,ierr)
    call MPI_COMM_RANK(comm,myrank,ierr)
    allocate(r_count(0:nproc-1),r_displ(0:nproc-1),stat=ierr)
    call chkalloc('evalxcsp','ptmp1',ierr)
    r_count = nlxyz/nproc
    do irnk = 0, mod(nlxyz,nproc)-1
       r_count(irnk) = r_count(irnk)+1
    end do
    r_displ(0) = 0
    do irnk = 0, nproc-2
       r_displ(irnk+1) = r_displ(irnk)+r_count(irnk)
    end do
    pnlxyz = r_count(myrank)
    allocate(prhop(pnlxyz,2),ppfxcn(pnlxyz,2),pexcor(pnlxyz),stat=ierr)
    call chkalloc('evalxcsp','ptmp2',ierr)
    !
!$omp parallel do private(j)
    do k=1,pnlxyz
       j = r_displ(myrank)+k

       prhop(k,1) = rhop(j,1)
       prhop(k,2) = rhop(j,2)
    end do
    select case(xctp)
    case (xcf_ldapw92)
       call ldapw92_exchcorsp(pnlxyz,prhop,pexc,pexcor,ppfxcn,mssg)
    case default
       if (mssg.gt.0) then
          write(6,*) 'evalxcsp(): unsupported value in xctp = ', xctp
       end if
       call  MPI_ABORT(MPI_COMM_WORLD,1,ierr)
       stop
    end select
    !
    call tpf_begin(tpf_mpi)
    call MPI_ALLREDUCE(pexc,exc,1,MPI_DOUBLE_PRECISION,MPI_SUM, &
         comm,ierr)
    call tpf_end(tpf_mpi)
    deallocate(prhop)
    !
    evxc(1:6) = 0.0d0

    allocate(pfxcn(nlxyz,2),stat=ierr)
    call chkalloc('evalxcsp','tmp3',ierr)
    call tpf_begin(tpf_mpi)
    call MPI_ALLGATHERV(ppfxcn(1,1),pnlxyz,MPI_DOUBLE_PRECISION, &
         pfxcn(1,1),r_count,r_displ,MPI_DOUBLE_PRECISION,comm,ierr)
    call MPI_ALLGATHERV(ppfxcn(1,2),pnlxyz,MPI_DOUBLE_PRECISION, &
         pfxcn(1,2),r_count,r_displ,MPI_DOUBLE_PRECISION,comm,ierr)

    call MPI_ALLGATHERV(pexcor,pnlxyz,MPI_DOUBLE_PRECISION, &
         excor,r_count,r_displ,MPI_DOUBLE_PRECISION,comm,ierr)
    call tpf_end(tpf_mpi)
    deallocate(ppfxcn,r_count,r_displ,pexcor)
    !
    do is=1,2
       call pad_zero(nrx2,nry2,nrz2,nlx,nly,nlz,pfxcn(1,is))
    end do

    ! generate Vxc(R) in rho
    call genvxcsp(nrx2*nry2*nrz2,nlxyz,pfxcn,rhop,rho,va, &
         exa,exc,evxc,evxcor(:,0))
    deallocate(pfxcn)
    return
  end subroutine evalxcsp
  !
  subroutine pad_zero(nx,ny,nz,lx,ly,lz,dat)
    implicit none
    integer,intent(in):: nx,ny,nz,lx,ly,lz
    real(kind=8):: dat(lx,ly,lz)

    integer:: j
    do j=nz+1,lz
       dat(:,:,j) = 0.0d0
    end do
    do j=nx+1,lx
       dat(j,:,:) = 0.0d0
    end do
    do j=ny+1,ly
       dat(:,j,:) = 0.0d0
    end do
    return
  end subroutine pad_zero
  !
  subroutine genvxc(nrxyz,nlxyz,part1,rhop,rho,va,exa,exc,evxc,evxc0or)
    implicit none
    integer:: nrxyz,nlxyz
    real(kind=8):: part1(nlxyz),rhop(nlxyz),rho(nlxyz)
    real(kind=8):: va,exa,exc,evxc(0:6)
    real(kind=8):: evxc0or(nlxyz)
    !
    integer:: i
    !
    evxc(0) = 0.0d0
    exa = exc
!$omp parallel do reduction(+:evxc,exa)
    do i=1,nlxyz
       evxc(0) = evxc(0) + part1(i)*rhop(i)
       exa = exa - part1(i)*rho(i)
       rho(i) = part1(i)
       evxc0or(i) = part1(i)*rhop(i)
    end do
    exa = exa*va/dble(nrxyz)
    exc = exc*va/dble(nrxyz)
    evxc(0) = evxc(0)*va/dble(nrxyz)
    return
  end subroutine genvxc
  !
  subroutine genvxcsp(nrxyz,nlxyz,part1,rhop,rho,va,exa,exc,evxc,evxc0or)
    implicit none
    integer:: nrxyz,nlxyz
    real(kind=8):: part1(nlxyz,2),rhop(nlxyz,2),rho(nlxyz,2)
    real(kind=8):: va,exa,exc,evxc(0:6)
    real(kind=8):: evxc0or(nlxyz)
    !
    integer:: i
    !
    evxc(0) = 0.0d0
    exa = exc
!$omp parallel do reduction(+:evxc,exa)
    do i=1,nlxyz
       evxc(0) = evxc(0) + rhop(i,1)*part1(i,1) + rhop(i,2)*part1(i,2)
       exa = exa - rho(i,1)*part1(i,1) - rho(i,2)*part1(i,2)
       rho(i,1) = part1(i,1)
       rho(i,2) = part1(i,2)
       evxc0or(i) = rhop(i,1)*part1(i,1) + rhop(i,2)*part1(i,2)
    end do
    exa = exa*va/dble(nrxyz)
    exc = exc*va/dble(nrxyz)
    evxc(0) = evxc(0)*va/dble(nrxyz)
    return
  end subroutine genvxcsp
end module pot_ldaxc
