!
! Copyright (c) 2017 Yoshihide Yoshimoto
!
#include "fftgrid.h"

!
! calculate energy density of and stress density of Hartree term.
!
! mode 1:
! \phi(r) = \int dr' \rho(r')/|r-r'| 
! E_hartree(r) = 1/2 \rho(r)\phi(r)
!
! mode 2:
! E(r) = -\grad\phi(r)
! \div E(r) = 4\pi\rho(r)
! E_hartree(r) = 1/(2*4\pi) E(r)*E(r)
!
module subr_shrtor
contains
  subroutine calc_shrtor(bb,mgs,nspin,ndens,rho,hrt,hrtor,shrtor,mode)
    use meshgrid
    use fft_3d
    implicit none
    real(kind=8),intent(in):: bb(3,3)
    type (meshgrid_struct):: mgs
    integer:: nspin,ndens
    real(kind=8),intent(in):: rho(2*mgs%nrx,2*mgs%nry,2*mgs%nrz,nspin*ndens)
    complex(kind=8),intent(in):: hrt(mgs%nkfxa)
    real(kind=8):: hrtor(2*mgs%nrx,2*mgs%nry,2*mgs%nrz)
    real(kind=8):: shrtor(2*mgs%nrx,2*mgs%nry,2*mgs%nrz,6)
    integer,intent(in):: mode

    if (mode.eq.1) then
       call calc_shrtor_core1(bb,mgs,nspin,ndens,rho,hrtor,shrtor)
    else
       call calc_shrtor_core2(bb,mgs,nspin,ndens,rho,hrtor,shrtor)
    end if
    return
  end subroutine calc_shrtor
  
  subroutine calc_shrtor_core1(bb,mgs,nspin,ndens,rho,hrtor,shrtor)
    use meshgrid
    use fft_3d
    implicit none
    real(kind=8),intent(in):: bb(3,3)
    type (meshgrid_struct):: mgs
    integer:: nspin,ndens
    real(kind=8),intent(in):: rho(2*mgs%nrx,2*mgs%nry,2*mgs%nrz,nspin*ndens)
    real(kind=8):: hrtor(2*mgs%nrx,2*mgs%nry,2*mgs%nrz)
    real(kind=8):: shrtor(2*mgs%nrx,2*mgs%nry,2*mgs%nrz,6)

    type (fft3_struct):: fs
    integer:: nrx2,nry2,nrz2
    real(kind=8),allocatable:: ztmp(:,:,:,:),wtmp(:,:,:,:),etmp(:,:,:,:,:)
    integer:: nkfxa
    integer,pointer:: kf(:,:)
    integer:: ig,k1,k2,k3,i1,i2,i3
    real(kind=8):: x1,x2,x3,fac
    real(kind=8),parameter:: pi = 3.1415926535897932385d0
    real(kind=8):: t1,t2,t3,tsq,tr,ti

    nkfxa = mgs%nkfxa
    kf => mgs%kf
    nrx2 = 2*mgs%nrx
    nry2 = 2*mgs%nry
    nrz2 = 2*mgs%nrz

    allocate(FFT_A3D_DEFSZ(ztmp,nrx2,nry2,nrz2), &
         FFT_A3D_DEFSZ(wtmp,nrx2,nry2,nrz2), &
         FFT_A3M_DEFSZ(etmp,nrx2,nry2,nrz2,7))

    call fft3_init(nrx2,nry2,nrz2,nrx2,nry2,nrz2,fs)

    ztmp = 0.0d0
    if (nspin.eq.1) then
       FFT_A3D_RL(ztmp,:,:,:) = rho(:,:,:,1)
    else
       FFT_A3D_RL(ztmp,:,:,:) = rho(:,:,:,1)+rho(:,:,:,2)
    end if

    call fft3_fw(fs,ztmp,wtmp)

    etmp = 0.0d0
!$omp parallel do private(ig,k1,k2,k3,x1,x2,x3,i1,i2,i3), &
!$omp private(fac,t1,t2,t3,tsq,tr,ti)
    do ig=2,nkfxa
       k1 = kf(1,ig)
       k2 = kf(2,ig)
       k3 = kf(3,ig)
       x1 = bb(1,1)*k1 + bb(1,2)*k2 + bb(1,3)*k3
       x2 = bb(2,1)*k1 + bb(2,2)*k2 + bb(2,3)*k3
       x3 = bb(3,1)*k1 + bb(3,2)*k2 + bb(3,3)*k3
       i1 = mod(k1+nrx2,nrx2)+1
       i2 = mod(k2+nry2,nry2)+1
       i3 = mod(k3+nrz2,nrz2)+1
       t1 = x1*x1
       t2 = x2*x2
       t3 = x3*x3
       tsq = t1+t2+t3
       fac = (4.0d0*pi)/tsq

       tr = fac*FFT_A3D_RL(ztmp,i1,i2,i3)
       ti = fac*FFT_A3D_IM(ztmp,i1,i2,i3)

       FFT_A3M_RL(etmp,i1,i2,i3,7) = tr
       FFT_A3M_IM(etmp,i1,i2,i3,7) = ti
       
       FFT_A3M_RL(etmp,i1,i2,i3,1) = (t1/tsq)*tr
       FFT_A3M_IM(etmp,i1,i2,i3,1) = (t1/tsq)*ti

       FFT_A3M_RL(etmp,i1,i2,i3,2) = (t2/tsq)*tr
       FFT_A3M_IM(etmp,i1,i2,i3,2) = (t2/tsq)*ti

       FFT_A3M_RL(etmp,i1,i2,i3,3) = (t3/tsq)*tr
       FFT_A3M_IM(etmp,i1,i2,i3,3) = (t3/tsq)*ti

       FFT_A3M_RL(etmp,i1,i2,i3,4) = (2.0d0*x2*x3/tsq)*tr
       FFT_A3M_IM(etmp,i1,i2,i3,4) = (2.0d0*x2*x3/tsq)*ti

       FFT_A3M_RL(etmp,i1,i2,i3,5) = (2.0d0*x3*x1/tsq)*tr
       FFT_A3M_IM(etmp,i1,i2,i3,5) = (2.0d0*x3*x1/tsq)*ti

       FFT_A3M_RL(etmp,i1,i2,i3,6) = (2.0d0*x1*x2/tsq)*tr
       FFT_A3M_IM(etmp,i1,i2,i3,6) = (2.0d0*x1*x2/tsq)*ti
    end do

    call fft3_bw(fs,FFT_A3M_RL(etmp,1,1,1,1),wtmp)
    call fft3_bw(fs,FFT_A3M_RL(etmp,1,1,1,2),wtmp)
    call fft3_bw(fs,FFT_A3M_RL(etmp,1,1,1,3),wtmp)
    call fft3_bw(fs,FFT_A3M_RL(etmp,1,1,1,4),wtmp)
    call fft3_bw(fs,FFT_A3M_RL(etmp,1,1,1,5),wtmp)
    call fft3_bw(fs,FFT_A3M_RL(etmp,1,1,1,6),wtmp)
    call fft3_bw(fs,FFT_A3M_RL(etmp,1,1,1,7),wtmp)

    if (nspin.eq.1) then
!$omp parallel do private(i1,i2,i3)
       do i3=1,nrz2
          do i2=1,nry2
             do i1=1,nrx2
                hrtor(i1,i2,i3) = 0.5d0*FFT_A3M_RL(etmp,i1,i2,i3,7) &
                     *rho(i1,i2,i3,1)
                shrtor(i1,i2,i3,1:6) = -FFT_A3M_RL(etmp,i1,i2,i3,1:6) &
                     *rho(i1,i2,i3,1)
             end do
          end do
       end do
    else
!$omp parallel do private(i1,i2,i3)
       do i3=1,nrz2
          do i2=1,nry2
             do i1=1,nrx2
                hrtor(i1,i2,i3) = 0.5d0*FFT_A3M_RL(etmp,i1,i2,i3,4) &
                     *( rho(i1,i2,i3,1) + rho(i1,i2,i3,2) )
                shrtor(i1,i2,i3,1:6) = -FFT_A3M_RL(etmp,i1,i2,i3,1:6) &
                     *( rho(i1,i2,i3,1) + rho(i1,i2,i3,2) )
             end do
          end do
       end do
    end if

    call fft3_done(fs)

    deallocate(ztmp,wtmp,etmp)
    return
  end subroutine calc_shrtor_core1

  subroutine calc_shrtor_core2(bb,mgs,nspin,ndens,rho,hrtor,shrtor)
    use meshgrid
    use fft_3d
    implicit none
    real(kind=8),intent(in):: bb(3,3)
    type (meshgrid_struct):: mgs
    integer:: nspin,ndens
    real(kind=8),intent(in):: rho(2*mgs%nrx,2*mgs%nry,2*mgs%nrz,nspin*ndens)
    real(kind=8):: hrtor(2*mgs%nrx,2*mgs%nry,2*mgs%nrz)
    real(kind=8):: shrtor(2*mgs%nrx,2*mgs%nry,2*mgs%nrz,6)

    type (fft3_struct):: fs
    integer:: nrx2,nry2,nrz2
    real(kind=8),allocatable:: ztmp(:,:,:,:),wtmp(:,:,:,:),etmp(:,:,:,:,:)
    integer:: nkfxa
    integer,pointer:: kf(:,:)
    integer:: ig,k1,k2,k3,i1,i2,i3
    real(kind=8):: x1,x2,x3,fac
    real(kind=8),parameter:: pi = 3.1415926535897932385d0
    real(kind=8):: t1,t2,t3

    nkfxa = mgs%nkfxa
    kf => mgs%kf
    nrx2 = 2*mgs%nrx
    nry2 = 2*mgs%nry
    nrz2 = 2*mgs%nrz

    allocate(FFT_A3D_DEFSZ(ztmp,nrx2,nry2,nrz2), &
         FFT_A3D_DEFSZ(wtmp,nrx2,nry2,nrz2), &
         FFT_A3M_DEFSZ(etmp,nrx2,nry2,nrz2,3))

    call fft3_init(nrx2,nry2,nrz2,nrx2,nry2,nrz2,fs)

    ztmp = 0.0d0
    if (nspin.eq.1) then
       FFT_A3D_RL(ztmp,:,:,:) = rho(:,:,:,1)
    else
       FFT_A3D_RL(ztmp,:,:,:) = rho(:,:,:,1)+rho(:,:,:,2)
    end if

    call fft3_fw(fs,ztmp,wtmp)

    etmp = 0.0d0
!$omp parallel do private(ig,k1,k2,k3,x1,x2,x3,i1,i2,i3,fac)
    do ig=2,nkfxa
       k1 = kf(1,ig)
       k2 = kf(2,ig)
       k3 = kf(3,ig)
       x1 = bb(1,1)*k1 + bb(1,2)*k2 + bb(1,3)*k3
       x2 = bb(2,1)*k1 + bb(2,2)*k2 + bb(2,3)*k3
       x3 = bb(3,1)*k1 + bb(3,2)*k2 + bb(3,3)*k3
       i1 = mod(k1+nrx2,nrx2)+1
       i2 = mod(k2+nry2,nry2)+1
       i3 = mod(k3+nrz2,nrz2)+1

       fac = (4.0d0*pi)/(x1*x1+x2*x2+x3*x3)
       FFT_A3M_RL(etmp,i1,i2,i3,1) = x1*fac*FFT_A3D_IM(ztmp,i1,i2,i3)
       FFT_A3M_IM(etmp,i1,i2,i3,1) = -x1*fac*FFT_A3D_RL(ztmp,i1,i2,i3)
       FFT_A3M_RL(etmp,i1,i2,i3,2) = x2*fac*FFT_A3D_IM(ztmp,i1,i2,i3)
       FFT_A3M_IM(etmp,i1,i2,i3,2) = -x2*fac*FFT_A3D_RL(ztmp,i1,i2,i3)
       FFT_A3M_RL(etmp,i1,i2,i3,3) = x3*fac*FFT_A3D_IM(ztmp,i1,i2,i3)
       FFT_A3M_IM(etmp,i1,i2,i3,3) = -x3*fac*FFT_A3D_RL(ztmp,i1,i2,i3)
    end do

    call fft3_bw(fs,FFT_A3M_RL(etmp,1,1,1,1),wtmp)
    call fft3_bw(fs,FFT_A3M_RL(etmp,1,1,1,2),wtmp)
    call fft3_bw(fs,FFT_A3M_RL(etmp,1,1,1,3),wtmp)

    fac = 1.0d0/(4.0d0*pi)
!$omp parallel do private(i1,i2,i3,t1,t2,t3)
    do i3=1,nrz2
       do i2=1,nry2
          do i1=1,nrx2
             t1 = FFT_A3M_RL(etmp,i1,i2,i3,1)*FFT_A3M_RL(etmp,i1,i2,i3,1)
             t2 = FFT_A3M_RL(etmp,i1,i2,i3,2)*FFT_A3M_RL(etmp,i1,i2,i3,2)
             t3 = FFT_A3M_RL(etmp,i1,i2,i3,3)*FFT_A3M_RL(etmp,i1,i2,i3,3)

             shrtor(i1,i2,i3,1) = -fac*t1
             shrtor(i1,i2,i3,2) = -fac*t2
             shrtor(i1,i2,i3,3) = -fac*t3
             shrtor(i1,i2,i3,4) = -fac*2.0d0*FFT_A3M_RL(etmp,i1,i2,i3,2) &
                  *FFT_A3M_RL(etmp,i1,i2,i3,3)
             shrtor(i1,i2,i3,5) = -fac*2.0d0*FFT_A3M_RL(etmp,i1,i2,i3,3) &
                  *FFT_A3M_RL(etmp,i1,i2,i3,1)
             shrtor(i1,i2,i3,6) = -fac*2.0d0*FFT_A3M_RL(etmp,i1,i2,i3,1) &
                  *FFT_A3M_RL(etmp,i1,i2,i3,2)
             hrtor(i1,i2,i3) = 0.5d0*fac*(t1 + t2 + t3)
          end do
       end do
    end do

    call fft3_done(fs)

    deallocate(ztmp,wtmp,etmp)
    return
  end subroutine calc_shrtor_core2

end module subr_shrtor
